/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  16.10.2016  mifi  First Version.
*  19.10.2016  mifi  Supports now SERCOM0, SERCOM1 and SERCOM4.
*  13.01.2018  mifi  Added supports for SERCOM3 (PA22,PA23) too.
**************************************************************************/
#define __TALCPU_COM_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <stdint.h>
#include "tal.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static TAL_COM_DCB *DCBArray[TAL_COM_PORT_MAX];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*
 * Calculate 64 bit division, ref can be found in
 * http://en.wikipedia.org/wiki/Division_algorithm#Long_division
 * (Based on sercom.c from xdk-asf-3.31.0)
 */
static uint64_t long_division (uint64_t n, uint64_t d)
{
   int32_t i;
   uint64_t q = 0, r = 0, bit_shift;
   for (i = 63; i >= 0; i--) {
      bit_shift = (uint64_t)1 << i;

      r = r << 1;

      if (n & bit_shift) {
         r |= 0x01;
      }

      if (r >= d) {
         r = r - d;
         q |= bit_shift;
      }
   }

   return q;
} /* long_division */ 

/*************************************************************************/
/*  WaitBusy                                                             */
/*                                                                       */
/*  Wait for the synchronization of registers between the                */ 
/*  clock domains is completed.                                          */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void WaitBusy (SercomUsart *pUSARTx)
{
   while(pUSARTx->SYNCBUSY.reg != 0)
   {
      __asm__ volatile ("nop");
   }
} /* WaitBusy */

/*************************************************************************/
/*  IRQHandler                                                           */
/*                                                                       */
/*  This is the generic IRQ handler.                                     */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void IRQHandler (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error;
   SercomUsart *pUSARTx;
   uint8_t      bStatus;
   uint8_t      bData;

   /* Get USART pointer */
   pUSARTx = (SercomUsart*)pDCB->HW.dBaseAddress;

   /* Get IRQ status */   
   bStatus = pUSARTx->INTFLAG.reg;
   
   /* Clear IRQ */
   pUSARTx->INTFLAG.reg = bStatus;

   /* 
    * Check for RX interrupt
    */
   if (bStatus & SERCOM_USART_INTFLAG_RXC)
   {
      /* Read data */
      bData = (uint8_t)pUSARTx->DATA.reg;
      
      /* If we have no overflow... */
      if (TAL_FALSE == pDCB->bRxOverflow)
      {
         /* ... put it into the ring buffer */
         Error = tal_MISCRingAdd(&pDCB->RxRing, &bData);
         if (TAL_OK == Error)
         {
            /* Signal counting semaphore */
            OSSemaSignalFromInt(&pDCB->RxRdySema);
         }
         else
         {
            /* Ups, overflow */
            pDCB->bRxOverflow = TAL_OK;
         }
      }  
   } /* end "RX interrupt" */
   
   
   /* 
    * Check for TX interrupt, but only if it is anabled 
    */
   if ( (pUSARTx->INTENSET.reg & SERCOM_USART_INTENSET_TXC) && (bStatus & SERCOM_USART_INTFLAG_TXC) )
   {
      /* Read Data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (Error != TAL_OK)
      {
         /* Ups, no data available, disable TX, enable RX interrupt */
         pUSARTx->INTENCLR.reg = SERCOM_USART_INTENCLR_MASK;   /* Disable all */
         pUSARTx->INTENSET.reg = SERCOM_USART_INTENSET_RXC;    /* Enable RX */
      }
      else
      {
         /* Send data */
         pUSARTx->DATA.reg = bData;
      }   
   } /* end "TX interrupt */
   
} /* IRQHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  SERCOMx_IRQHandler                                                   */
/*                                                                       */
/*  This is the Cortex SERCOMx IRQ handler.                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void SERCOM0_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_1]);
   SERCOM0_IRQHandler();
} /* SERCOM0_IRQHandler */

void SERCOM1_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_2]);
   SERCOM1_IRQHandler();
} /* SERCOM1_IRQHandler */

void SERCOM2_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_3]);
   TAL_CPU_IRQ_EXIT();
} /* SERCOM2_IRQHandler */

void SERCOM3_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_4]);
   TAL_CPU_IRQ_EXIT();
} /* SERCOM3_IRQHandler */

void SERCOM4_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_5]);
   TAL_CPU_IRQ_EXIT();
} /* SERCOM4_IRQHandler */

void SERCOM5_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_6]);
   TAL_CPU_IRQ_EXIT();
} /* SERCOM5_IRQHandler */

/*************************************************************************/
/*  cpu_COMInit                                                          */
/*                                                                       */
/*  Prepare the hardware for use by the Open function later. Set the HW  */
/*  information depending of ePort and "enable" the COM port.            */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMInit (TAL_COM_DCB *pDCB)
{
   TAL_RESULT   Error = TAL_ERR_COM_PORT_RANGE;
   TAL_COM_HW *pHW    = &pDCB->HW;
   
   switch (pDCB->ePort)
   {
      case TAL_COM_PORT_1:
      {
         Error = tal_BoardEnableCOM1();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_1] = pDCB;
         
            pHW->dBaseAddress = SERCOM0_BASE;
            pHW->IrqNumber    = SERCOM0_IRQn;
            pHW->bIrqPriority = SERCOM0_PRIO;            
         }
         break;
      } /* TAL_COM_PORT_1 */
   
      case TAL_COM_PORT_2:
      {
         Error = tal_BoardEnableCOM2();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_2] = pDCB;
         
            pHW->dBaseAddress = SERCOM1_BASE;
            pHW->IrqNumber    = SERCOM1_IRQn;
            pHW->bIrqPriority = SERCOM1_PRIO;            
         }
         break;
      } /* TAL_COM_PORT_2 */
   
      case TAL_COM_PORT_4:
      {
         Error = tal_BoardEnableCOM4();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_4] = pDCB;
         
            pHW->dBaseAddress = SERCOM3_BASE;
            pHW->IrqNumber    = SERCOM3_IRQn;
            pHW->bIrqPriority = SERCOM3_PRIO;            
         }
         break;
      } /* TAL_COM_PORT_4 */
   
      case TAL_COM_PORT_5:
      {
         Error = tal_BoardEnableCOM5();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_5] = pDCB;
         
            pHW->dBaseAddress = SERCOM4_BASE;
            pHW->IrqNumber    = SERCOM4_IRQn;
            pHW->bIrqPriority = SERCOM4_PRIO;            
         }
         break;
      } /* TAL_COM_PORT_5 */
   
      default:
      {
         /* Do nothing */
         break;
      }
   } /* end switch (pDCB->ePort) */
   
   return(Error);
} /* cpu_COMInit */

/*************************************************************************/
/*  cpu_COMOpen                                                          */
/*                                                                       */
/*  Open the COM port.                                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMOpen (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_ERROR;
   SercomUsart *pUSARTx;
   uint32_t     dCtrlA;
   uint32_t     dCtrlB;
   uint16_t     wBaud;
   uint8_t      bDummy;

   /* Get USART pointer */
   pUSARTx = (SercomUsart *)pDCB->HW.dBaseAddress;

   dCtrlA = pUSARTx->CTRLA.reg;
   dCtrlB = pUSARTx->CTRLB.reg;

   /* Check word length */
   switch (pDCB->Settings.eLength)
   {
      case TAL_COM_LENGTH_8:
      {
         dCtrlB &= ~SERCOM_USART_CTRLB_CHSIZE_Msk;
         break;
      }
      case TAL_COM_LENGTH_9:
      {
         dCtrlB |= SERCOM_USART_CTRLB_CHSIZE(1);
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_LENGTH;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eLength) */
   
   /* Check parity settings */
   switch (pDCB->Settings.eParity)
   {
      case TAL_COM_PARITY_NONE:
      {
         /* Do nothing */
         break;
      } 
      
      case TAL_COM_PARITY_EVEN:
      {
         dCtrlA |= SERCOM_USART_CTRLA_FORM(1);
         break;
      }
      
      case TAL_COM_PARITY_ODD:
      {
         dCtrlA |= SERCOM_USART_CTRLA_FORM(1);
         dCtrlB |= SERCOM_USART_CTRLB_PMODE;
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_PARITY;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eParity) */
    
   /* Check stop bit settings */
   switch (pDCB->Settings.eStop)
   {
      case TAL_COM_STOP_1_0:
      {
         /* Do nothing */
         break;
      }
      
      case TAL_COM_STOP_2_0:
      {
         dCtrlB |= SERCOM_USART_CTRLB_SBMODE; 
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_STOP;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eStop) */

   /* Check baud rate */
   if ((pDCB->Settings.dBaudrate * 16) > tal_CPUGetFrequencyCPU())
   {
      Error = TAL_ERR_COM_BAUDRATE;
      goto COMOpenEnd;  /*lint !e801*/
   }

   if (pDCB->Settings.dBaudrate != 0)
   {
      uint64_t temp1;
      uint64_t ratio;
      uint64_t scale;
      
      temp1 = ((16 * (uint64_t)pDCB->Settings.dBaudrate) << 32);
      ratio = long_division(temp1, (uint64_t)tal_CPUGetFrequencyCPU());
      scale = ((uint64_t)1 << 32) - ratio;
      wBaud = (uint16_t)((65536 * scale) >> 32);
   }
   else
   {
      Error = TAL_ERR_COM_BAUDRATE;
      goto COMOpenEnd;  /*lint !e801*/
   }

   /*
    * Set USART register
    */

   /* Set BAUD register */    
   pUSARTx->BAUD.reg  = wBaud;

   /* Enable transmitter and receiver */
   dCtrlB |= (SERCOM_USART_CTRLB_RXEN | SERCOM_USART_CTRLB_TXEN);
   
   /* USART with internal clock */
   dCtrlA |= SERCOM_USART_CTRLA_MODE_USART_INT_CLK;

   /* LSB is transmitted first. */
   dCtrlA |= SERCOM_USART_CTRLA_DORD;

   /* Enable USART itself */
   dCtrlA |= SERCOM_USART_CTRLA_ENABLE;

   /* Set CTRLB register */
   pUSARTx->CTRLB.reg = dCtrlB;
   WaitBusy(pUSARTx);

   /* Set CTRLA register */
   pUSARTx->CTRLA.reg = dCtrlA;
   WaitBusy(pUSARTx);

   /* Dummy read to clear the DR register */
   bDummy = (uint8_t)pUSARTx->DATA.reg;
   (void)bDummy;

   /* Enable RX interrrupt */
   pUSARTx->INTENSET.reg = SERCOM_USART_INTENSET_RXC;
   
   /* Set priority and enable the IRQ */
   tal_CPUIrqSetPriority(pDCB->HW.IrqNumber, pDCB->HW.bIrqPriority);
   tal_CPUIrqEnable(pDCB->HW.IrqNumber);

   Error = TAL_OK;
   
COMOpenEnd:   
   return(Error);
} /* cpu_COMOpen */

/*************************************************************************/
/*  cpu_COMClose                                                         */
/*                                                                       */
/*  Close the COM port.                                                  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMClose (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   SercomUsart *pUSARTx;
   
   /* Get USART pointer */
   pUSARTx = (SercomUsart*)pDCB->HW.dBaseAddress;

   /* Disable the IRQ */   
   tal_CPUIrqDisable(pDCB->HW.IrqNumber);
   
   /* Disable the USART itself */
   pUSARTx->CTRLA.reg &= ~SERCOM_USART_CTRLA_ENABLE;
   WaitBusy(pUSARTx);
   
   /* Disable USART interrupts */   
   pUSARTx->INTENCLR.reg = 0x0F;
   
   return(Error);
} /* cpu_COMClose */

/*************************************************************************/
/*  cpu_COMStartTx                                                       */
/*                                                                       */
/*  Send the data from the ring buffer if the TX interrupt is disabled.  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMStartTx (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   SercomUsart *pUSARTx;
   uint8_t      bData;
   uint8_t      bValue;
   
   /* Get USART pointer */
   pUSARTx = (SercomUsart*)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   bValue = pUSARTx->INTENSET.reg & SERCOM_USART_INTENSET_MASK;
   if (bValue & SERCOM_USART_INTENSET_TXC)
   {
      /* TX interrupt is enabled, do nothing */
   }
   else
   {
      /* Get data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (TAL_OK == Error)
      {
         /* Send data */
         pUSARTx->DATA.reg = bData;
      
         /* Enable TX interrupt */
         bValue |= SERCOM_USART_INTENSET_TXC;
         pUSARTx->INTENSET.reg = bValue;
      }
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMStartTx */

/*************************************************************************/
/*  cpu_COMTxIsRunning                                                   */
/*                                                                       */
/*  Check if TX is still running.                                        */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT cpu_COMTxIsRunning (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   SercomUsart *pUSARTx;
   uint8_t      bValue;
   
   /* Get USART pointer */
   pUSARTx = (SercomUsart*)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   bValue = pUSARTx->INTENSET.reg & SERCOM_USART_INTENSET_MASK;
   if (bValue & SERCOM_USART_INTENSET_TXC)
   {
      /* TX is still running */
      Error = TAL_OK;
   }
   else
   {
      /* TX is not running */
      Error = TAL_ERROR;
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMTxIsRunning */

/*** EOF ***/
